@@ -15,14 +15,15 @@
  * License along with this library; if not, see <https://www.gnu.org/licenses/>.
  */
 
-#include "shumate-vector-renderer.h"
+#include "shumate-vector-renderer-private.h"
 #include "shumate-tile-downloader.h"
 #include "shumate-tile-private.h"
+#include "shumate-profiling-private.h"
 
 /**
  * ShumateVectorRenderer:
  *
- * A [class@Shumate.MapSource] that renders tiles from a given vector data source.
+ * A [class@MapSource] that renders tiles from a given vector data source.
  */
 
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
@@ -30,6 +31,7 @@
 #include <cairo/cairo.h>
 
 #include "vector/shumate-vector-render-scope-private.h"
+#include "vector/shumate-vector-symbol-info-private.h"
 #include "vector/shumate-vector-utils-private.h"
 #include "vector/shumate-vector-layer-private.h"
 #endif
@@ -40,17 +42,23 @@ struct _ShumateVectorRenderer
 
   char *source_name;
   ShumateDataSource *data_source;
-  GPtrArray *tiles;
 
-#ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorSpriteSheet *sprites;
-#endif
+  GMutex sprites_mutex;
+
+  GThreadPool *thread_pool;
 
   char *style_json;
 
   GPtrArray *layers;
 };
 
+
+static gboolean begin_render (ShumateVectorRenderer  *self,
+                              GTask                  *task,
+                              GBytes                 *tile_data,
+                              ShumateGridPosition    *source_position);
+
 static void shumate_vector_renderer_initable_iface_init (GInitableIface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (ShumateVectorRenderer, shumate_vector_renderer, SHUMATE_TYPE_MAP_SOURCE,
@@ -59,6 +67,7 @@ G_DEFINE_TYPE_WITH_CODE (ShumateVectorRenderer, shumate_vector_renderer, SHUMATE
 enum {
   PROP_0,
   PROP_STYLE_JSON,
+  PROP_SPRITE_SHEET,
   N_PROPS
 };
 
@@ -117,15 +126,6 @@ shumate_vector_renderer_is_supported (void)
 }
 
 
-static void
-on_data_source_received_data (ShumateVectorRenderer *self,
-                              int                    x,
-                              int                    y,
-                              int                    zoom_level,
-                              GBytes                *bytes,
-                              ShumateDataSource     *data_source);
-
-
 static void
 shumate_vector_renderer_finalize (GObject *object)
 {
@@ -135,10 +135,12 @@ shumate_vector_renderer_finalize (GObject *object)
   g_clear_pointer (&self->style_json, g_free);
   g_clear_pointer (&self->source_name, g_free);
   g_clear_object (&self->data_source);
-  g_clear_pointer (&self->tiles, g_ptr_array_unref);
-#ifdef SHUMATE_HAS_VECTOR_RENDERER
   g_clear_object (&self->sprites);
-#endif
+
+  if (self->thread_pool)
+    g_thread_pool_free (self->thread_pool, FALSE, FALSE);
+
+  g_mutex_clear (&self->sprites_mutex);
 
   G_OBJECT_CLASS (shumate_vector_renderer_parent_class)->finalize (object);
 }
@@ -156,6 +158,9 @@ shumate_vector_renderer_get_property (GObject    *object,
     case PROP_STYLE_JSON:
       g_value_set_string (value, self->style_json);
       break;
+    case PROP_SPRITE_SHEET:
+      g_value_set_object (value, shumate_vector_renderer_get_sprite_sheet (self));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -176,6 +181,9 @@ shumate_vector_renderer_set_property (GObject      *object,
       g_assert (self->style_json == NULL);
       self->style_json = g_strdup (g_value_get_string (value));
       break;
+    case PROP_SPRITE_SHEET:
+      shumate_vector_renderer_set_sprite_sheet (self, g_value_get_object (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -219,6 +227,20 @@ shumate_vector_renderer_class_init (ShumateVectorRendererClass *klass)
                          NULL,
                          G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * ShumateVectorRenderer:sprite-sheet:
+   *
+   * The sprite sheet used to render icons and textures.
+   *
+   * Since: 1.1
+   */
+  properties[PROP_SPRITE_SHEET] =
+    g_param_spec_object ("sprite-sheet",
+                         "sprite-sheet",
+                         "sprite-sheet",
+                         SHUMATE_TYPE_VECTOR_SPRITE_SHEET,
+                         G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
   g_object_class_install_properties (object_class, N_PROPS, properties);
 }
 
@@ -228,6 +250,8 @@ shumate_vector_renderer_initable_init (GInitable     *initable,
                                        GCancellable  *cancellable,
                                        GError       **error)
 {
+  SHUMATE_PROFILE_START ();
+
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorRenderer *self = (ShumateVectorRenderer *)initable;
   g_autoptr(JsonNode) node = NULL;
@@ -338,16 +362,19 @@ shumate_vector_renderer_initable_init (GInitable     *initable,
                                                error))
             return FALSE;
 
-          minzoom = MIN (minzoom, json_object_get_int_member_with_default (source_object, "minzoom", 0));
-          maxzoom = MAX (maxzoom, json_object_get_int_member_with_default (source_object, "maxzoom", 30));
+          minzoom = json_object_get_int_member_with_default (source_object, "minzoom", 0);
+          maxzoom = json_object_get_int_member_with_default (source_object, "maxzoom", 30);
 
           data_source = SHUMATE_DATA_SOURCE (shumate_tile_downloader_new (url_template));
-          g_signal_connect_object (data_source, "received-data", (GCallback)on_data_source_received_data, self, G_CONNECT_SWAPPED);
+          shumate_data_source_set_min_zoom_level (data_source, minzoom);
+          shumate_data_source_set_max_zoom_level (data_source, maxzoom);
 
           self->source_name = g_strdup (source_name);
           self->data_source = data_source;
         }
 
+      maxzoom = MAX (maxzoom, 18);
+
       if (minzoom < maxzoom)
         {
           shumate_map_source_set_min_zoom_level (SHUMATE_MAP_SOURCE (self), minzoom);
@@ -407,10 +434,53 @@ shumate_vector_renderer_initable_iface_init (GInitableIface *iface)
 }
 
 
+/* An item in the thread pool queue to render a tile from received data. */
+typedef struct {
+  GTask *task;
+  GCancellable *cancellable;
+  gulong cancellable_handle;
+  GBytes *data;
+  ShumateGridPosition source_position;
+
+  GdkPaintable *paintable;
+  GPtrArray *symbols;
+} RenderJob;
+
+
+static void
+render_job_unref (RenderJob *job)
+{
+  if (job->cancellable_handle != 0)
+    g_cancellable_disconnect (g_task_get_cancellable (job->task), job->cancellable_handle);
+  g_clear_object (&job->cancellable);
+  g_clear_object (&job->task);
+  g_clear_pointer (&job->data, g_bytes_unref);
+  g_clear_object (&job->paintable);
+  g_clear_pointer (&job->symbols, g_ptr_array_unref);
+  g_free (job);
+}
+
+/* The data associated with a shumate_vector_renderer_fill_tile_async() task. */
+typedef struct {
+  ShumateTile *tile;
+  RenderJob *current_job;
+  ShumateDataSourceRequest *req;
+  guint8 completed : 1;
+} TaskData;
+
+static void
+task_data_free (TaskData *data)
+{
+  g_clear_object (&data->tile);
+  g_clear_object (&data->req);
+  g_free (data);
+}
+
+
 static void
 shumate_vector_renderer_init (ShumateVectorRenderer *self)
 {
-  self->tiles = g_ptr_array_new_full (0, g_object_unref);
+  g_mutex_init (&self->sprites_mutex);
 }
 
 
@@ -434,36 +504,91 @@ shumate_vector_renderer_get_style_json (ShumateVectorRenderer *self)
 /**
  * shumate_vector_renderer_set_sprite_sheet_data:
  * @self: a [class@VectorRenderer]
- * @sprites_pixbuf: a [class@Gdk.Pixbuf]
+ * @sprites_pixbuf: a [class@GdkPixbuf.Pixbuf]
  * @sprites_json: a JSON string
  * @error: return location for a #GError, or %NULL
  *
  * Sets the sprite sheet used by the style JSON to render icons and textures.
  *
- * See <https://maplibre.org/maplibre-gl-js-docs/style-spec/sprite/> for
- * details about the spritesheet format. Most stylesheets provide these files
- * along with the main style JSON.
+ * The existing [property@VectorRenderer:sprite-sheet] property will be replaced
+ * with a new instance of [class@VectorSpriteSheet].
+ *
+ * Returns: whether the sprite sheet was loaded successfully
+ *
+ * Deprecated: 1.1: Use the methods of [property@VectorRenderer:sprite-sheet] instead.
  */
-void
+gboolean
 shumate_vector_renderer_set_sprite_sheet_data (ShumateVectorRenderer  *self,
                                                GdkPixbuf              *sprites_pixbuf,
                                                const char             *sprites_json,
                                                GError                **error)
 {
-  g_return_if_fail (SHUMATE_IS_VECTOR_RENDERER (self));
-  g_return_if_fail (GDK_IS_PIXBUF (sprites_pixbuf));
-  g_return_if_fail (sprites_json != NULL);
+  g_autoptr(ShumateVectorSpriteSheet) sprites = NULL;
+  g_autoptr(GdkTexture) texture = NULL;
 
-#ifdef SHUMATE_HAS_VECTOR_RENDERER
-  g_clear_object (&self->sprites);
-  self->sprites = shumate_vector_sprite_sheet_new (sprites_pixbuf, sprites_json, NULL, error);
-#else
-  g_set_error (error,
-               SHUMATE_STYLE_ERROR,
-               SHUMATE_STYLE_ERROR_SUPPORT_OMITTED,
-               "Libshumate was compiled without support for vector tiles.");
-  return FALSE;
-#endif
+  g_return_val_if_fail (SHUMATE_IS_VECTOR_RENDERER (self), FALSE);
+  g_return_val_if_fail (GDK_IS_PIXBUF (sprites_pixbuf), FALSE);
+  g_return_val_if_fail (sprites_json != NULL, FALSE);
+
+  sprites = shumate_vector_sprite_sheet_new ();
+
+  texture = gdk_texture_new_for_pixbuf (sprites_pixbuf);
+  if (!shumate_vector_sprite_sheet_add_page (self->sprites, texture, sprites_json, 1, error))
+    return FALSE;
+
+  shumate_vector_renderer_set_sprite_sheet (self, sprites);
+  return TRUE;
+}
+
+
+/**
+ * shumate_vector_renderer_get_sprite_sheet:
+ * @self: a [class@VectorRenderer]
+ *
+ * Gets the sprite sheet used to render icons and textures.
+ *
+ * Returns: (transfer none): the [class@VectorSpriteSheet]
+ *
+ * Since: 1.1
+ */
+ShumateVectorSpriteSheet *
+shumate_vector_renderer_get_sprite_sheet (ShumateVectorRenderer *self)
+{
+  g_autoptr(GMutexLocker) locker = NULL;
+
+  g_return_val_if_fail (SHUMATE_IS_VECTOR_RENDERER (self), FALSE);
+
+  locker = g_mutex_locker_new (&self->sprites_mutex);
+
+  if (self->sprites == NULL)
+    self->sprites = shumate_vector_sprite_sheet_new ();
+
+  return self->sprites;
+}
+
+
+/**
+ * shumate_vector_renderer_set_sprite_sheet:
+ * @self: a [class@VectorRenderer]
+ * @sprites: a [class@VectorSpriteSheet]
+ *
+ * Sets the sprite sheet used to render icons and textures.
+ *
+ * Since: 1.1
+ */
+void
+shumate_vector_renderer_set_sprite_sheet (ShumateVectorRenderer    *self,
+                                          ShumateVectorSpriteSheet *sprites)
+{
+  g_autoptr(GMutexLocker) locker = NULL;
+
+  g_return_if_fail (SHUMATE_IS_VECTOR_RENDERER (self));
+  g_return_if_fail (SHUMATE_IS_VECTOR_SPRITE_SHEET (sprites));
+
+  locker = g_mutex_locker_new (&self->sprites_mutex);
+
+  if (g_set_object (&self->sprites, sprites))
+    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SPRITE_SHEET]);
 }
 
 
@@ -495,9 +620,76 @@ texture_new_for_surface (cairo_surface_t *surface)
 #endif
 
 
-static void on_data_source_done (GObject      *object,
-                                 GAsyncResult *res,
-                                 gpointer      user_data);
+static void
+get_source_coordinates (ShumateVectorRenderer *self,
+                        int                   *x,
+                        int                   *y,
+                        int                   *zoom_level)
+{
+  /* Figure out which tile from the data source should be used to render the
+   * given tile (which will be different if we're overzooming). */
+
+  int maxzoom = shumate_data_source_get_max_zoom_level (self->data_source);
+  if (*zoom_level > maxzoom)
+    {
+      *x >>= (*zoom_level - maxzoom);
+      *y >>= (*zoom_level - maxzoom);
+      *zoom_level = maxzoom;
+    }
+}
+
+static void
+on_request_notify (ShumateDataSourceRequest *req,
+                   GParamSpec               *pspec,
+                   gpointer                  user_data)
+{
+  GTask *task = user_data;
+  GBytes *data;
+  ShumateVectorRenderer *self = g_task_get_source_object (task);
+
+  if ((data = shumate_data_source_request_get_data (req)))
+    {
+      begin_render (
+        self,
+        task,
+        data,
+        &SHUMATE_GRID_POSITION_INIT (
+          shumate_data_source_request_get_x (req),
+          shumate_data_source_request_get_y (req),
+          shumate_data_source_request_get_zoom_level (req)
+        )
+      );
+    }
+}
+
+
+static void
+return_from_task (GTask *task, ShumateDataSourceRequest *req)
+{
+  GError *error;
+  TaskData *data = g_task_get_task_data (task);
+
+  shumate_tile_set_state (data->tile, SHUMATE_STATE_DONE);
+
+  if ((error = shumate_data_source_request_get_error (req)))
+    g_task_return_error (task, g_error_copy (error));
+  else
+    g_task_return_boolean (task, TRUE);
+}
+
+static void
+on_request_notify_completed (ShumateDataSourceRequest *req,
+                             GParamSpec               *pspec,
+                             gpointer                  user_data)
+{
+  g_autoptr(GTask) task = user_data;
+  TaskData *data = g_task_get_task_data (task);
+
+  if (data->current_job != NULL)
+    data->completed = TRUE;
+  else
+    return_from_task (task, req);
+}
 
 static void
 shumate_vector_renderer_fill_tile_async (ShumateMapSource    *map_source,
@@ -508,6 +700,10 @@ shumate_vector_renderer_fill_tile_async (ShumateMapSource    *map_source,
 {
   ShumateVectorRenderer *self = (ShumateVectorRenderer *)map_source;
   g_autoptr(GTask) task = NULL;
+  int x, y, zoom_level;
+  g_autoptr(ShumateDataSourceRequest) req = NULL;
+  GBytes *data;
+  TaskData *task_data;
 
   g_return_if_fail (SHUMATE_IS_VECTOR_RENDERER (self));
   g_return_if_fail (SHUMATE_IS_TILE (tile));
@@ -515,17 +711,37 @@ shumate_vector_renderer_fill_tile_async (ShumateMapSource    *map_source,
 
   task = g_task_new (self, cancellable, callback, user_data);
   g_task_set_source_tag (task, shumate_vector_renderer_fill_tile_async);
-  g_task_set_task_data (task, g_object_ref (tile), (GDestroyNotify)g_object_unref);
 
-  g_ptr_array_add (self->tiles, g_object_ref (tile));
+  task_data = g_new0 (TaskData, 1);
+  task_data->tile = g_object_ref (tile);
+  g_task_set_task_data (task, task_data, (GDestroyNotify)task_data_free);
 
-  shumate_data_source_get_tile_data_async (self->data_source,
-                                           shumate_tile_get_x (tile),
-                                           shumate_tile_get_y (tile),
-                                           shumate_tile_get_zoom_level (tile),
-                                           cancellable,
-                                           on_data_source_done,
-                                           g_steal_pointer (&task));
+  x = shumate_tile_get_x (tile);
+  y = shumate_tile_get_y (tile);
+  zoom_level = shumate_tile_get_zoom_level (tile);
+
+  get_source_coordinates (self, &x, &y, &zoom_level);
+
+  req = shumate_data_source_start_request (self->data_source, x, y, zoom_level, cancellable);
+  task_data->req = g_object_ref (req);
+
+  if ((data = shumate_data_source_request_get_data (req)))
+    {
+      begin_render (
+        self,
+        task,
+        data,
+        &SHUMATE_GRID_POSITION_INIT (x, y, zoom_level)
+      );
+    }
+
+  if (shumate_data_source_request_is_completed (req))
+    on_request_notify_completed (req, NULL, g_steal_pointer (&task));
+  else
+    {
+      g_signal_connect_object (req, "notify::data", (GCallback)on_request_notify, task, 0);
+      g_signal_connect_object (req, "notify::completed", (GCallback)on_request_notify_completed, g_object_ref (task), 0);
+    }
 }
 
 static gboolean
@@ -541,97 +757,198 @@ shumate_vector_renderer_fill_tile_finish (ShumateMapSource  *map_source,
   return g_task_propagate_boolean (G_TASK (result), error);
 }
 
-static void
-on_data_source_done (GObject *object, GAsyncResult *res, gpointer user_data)
+void
+shumate_vector_renderer_render (ShumateVectorRenderer  *self,
+                                ShumateTile            *tile,
+                                GBytes                 *tile_data,
+                                ShumateGridPosition    *source_position,
+                                GdkPaintable          **paintable,
+                                GPtrArray             **symbols)
 {
-  ShumateDataSource *data_source = SHUMATE_DATA_SOURCE (object);
-  g_autoptr(GTask) task = G_TASK (user_data);
-  ShumateVectorRenderer *self = g_task_get_source_object (task);
-  ShumateTile *tile = g_task_get_task_data (task);
-  GError *error = NULL;
+  SHUMATE_PROFILE_START ();
 
-  g_ptr_array_remove_fast (self->tiles, tile);
-
-  if (!shumate_data_source_get_tile_data_finish (data_source, res, &error))
-    g_task_return_error (task, error);
-  else
-    {
-      shumate_tile_set_state (tile, SHUMATE_STATE_DONE);
-      g_task_return_boolean (task, TRUE);
-    }
-}
-
-static void
-render (ShumateVectorRenderer *self,
-        ShumateTile           *tile,
-        GBytes                *tile_data,
-        double                 zoom_level)
-{
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorRenderScope scope;
-  g_autoptr(GdkTexture) texture = NULL;
   cairo_surface_t *surface;
   gconstpointer data;
   gsize len;
-  g_autoptr(GPtrArray) symbols = g_ptr_array_new ();
+  g_autoptr(GPtrArray) symbol_list = g_ptr_array_new_with_free_func ((GDestroyNotify)shumate_vector_symbol_info_unref);
   int texture_size;
+  g_autofree char *profile_desc = NULL;
+  g_autoptr(ShumateVectorSpriteSheet) sprites = NULL;
 
   g_assert (SHUMATE_IS_VECTOR_RENDERER (self));
   g_assert (SHUMATE_IS_TILE (tile));
 
-  texture_size = shumate_tile_get_size (tile);
-  scope.target_size = texture_size;
-  scope.zoom_level = zoom_level;
-  scope.symbols = symbols;
-  scope.sprites = self->sprites;
+  g_mutex_lock (&self->sprites_mutex);
+  if (self->sprites == NULL)
+    self->sprites = shumate_vector_sprite_sheet_new ();
+  sprites = g_object_ref (self->sprites);
+  g_mutex_unlock (&self->sprites_mutex);
 
-  surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, texture_size, texture_size);
+  texture_size = shumate_tile_get_size (tile);
+  scope.scale_factor = shumate_tile_get_scale_factor (tile);
+  scope.target_size = texture_size;
+  scope.tile_x = shumate_tile_get_x (tile);
+  scope.tile_y = shumate_tile_get_y (tile);
+  scope.zoom_level = shumate_tile_get_zoom_level (tile);
+  scope.symbols = symbol_list;
+  scope.sprites = sprites;
+
+  if (scope.zoom_level > source_position->zoom)
+    {
+      float s = 1 << ((int)scope.zoom_level - source_position->zoom);
+      scope.overzoom_x = (scope.tile_x - (source_position->x << ((int)scope.zoom_level - source_position->zoom))) / s;
+      scope.overzoom_y = (scope.tile_y - (source_position->y << ((int)scope.zoom_level - source_position->zoom))) / s;
+      scope.overzoom_scale = s;
+    }
+  else
+    {
+      scope.overzoom_x = 0;
+      scope.overzoom_y = 0;
+      scope.overzoom_scale = 1;
+    }
+
+  surface = cairo_image_surface_create (
+    CAIRO_FORMAT_ARGB32,
+    texture_size * scope.scale_factor,
+    texture_size * scope.scale_factor
+  );
   scope.cr = cairo_create (surface);
+  cairo_scale (scope.cr, scope.scale_factor, scope.scale_factor);
 
   data = g_bytes_get_data (tile_data, &len);
   scope.tile = vector_tile__tile__unpack (NULL, len, data);
 
   if (scope.tile != NULL)
-    for (int i = 0; i < self->layers->len; i ++)
-      shumate_vector_layer_render ((ShumateVectorLayer *)self->layers->pdata[i], &scope);
+    for (scope.layer_idx = 0; scope.layer_idx < self->layers->len; scope.layer_idx ++)
+      shumate_vector_layer_render ((ShumateVectorLayer *)self->layers->pdata[scope.layer_idx], &scope);
 
-  texture = texture_new_for_surface (surface);
-  shumate_tile_set_paintable (tile, GDK_PAINTABLE (texture));
-  shumate_tile_set_symbols (tile, symbols);
+  *paintable = GDK_PAINTABLE (texture_new_for_surface (surface));
+  *symbols = g_ptr_array_ref (scope.symbols);
 
   cairo_destroy (scope.cr);
   cairo_surface_destroy (surface);
   vector_tile__tile__free_unpacked (scope.tile, NULL);
+
+  profile_desc = g_strdup_printf ("(%d, %d) @ %f", scope.tile_x, scope.tile_y, scope.zoom_level);
+  SHUMATE_PROFILE_END (profile_desc);
+
 #else
   g_return_if_reached ();
 #endif
 }
 
-static void
-on_data_source_received_data (ShumateVectorRenderer *self,
-                              int                    x,
-                              int                    y,
-                              int                    zoom_level,
-                              GBytes                *bytes,
-                              ShumateDataSource     *data_source)
+static gboolean
+render_job_finish (RenderJob *job)
 {
-  int i;
-  ShumateTile *tile;
+  TaskData *data = g_task_get_task_data (job->task);
 
-  g_assert (SHUMATE_IS_VECTOR_RENDERER (self));
-  g_assert (SHUMATE_IS_DATA_SOURCE (data_source));
-
-  for (i = 0; i < self->tiles->len; i ++)
+  if (!g_cancellable_is_cancelled (job->cancellable))
     {
-      tile = self->tiles->pdata[i];
-
-      if (shumate_tile_get_x (tile) == x
-          && shumate_tile_get_y (tile) == y
-          && shumate_tile_get_zoom_level (tile) == zoom_level)
-          render (self, tile, bytes, zoom_level);
+      shumate_tile_set_paintable (data->tile, job->paintable);
+      shumate_tile_set_symbols (data->tile, job->symbols);
     }
+
+  if (data->completed)
+    return_from_task (job->task, data->req);
+
+  /* Make sure the TaskData knows there's not a current job anymore */
+  if (data->current_job == job)
+    data->current_job = NULL;
+
+  render_job_unref (job);
+  return G_SOURCE_REMOVE;
 }
 
+static void
+thread_func (RenderJob *job)
+{
+  ShumateVectorRenderer *self = g_task_get_source_object (job->task);
+  TaskData *data = g_task_get_task_data (job->task);
+
+  if (!g_cancellable_is_cancelled (job->cancellable))
+    {
+      shumate_vector_renderer_render (
+        self,
+        data->tile,
+        job->data,
+        &job->source_position,
+        &job->paintable,
+        &job->symbols
+      );
+    }
+
+  g_idle_add ((GSourceFunc)render_job_finish, job);
+}
+
+static void
+chain_cancel (GCancellable *source,
+              GCancellable *dest)
+{
+  g_cancellable_cancel (dest);
+}
+
+static gboolean
+begin_render (ShumateVectorRenderer  *self,
+              GTask                  *task,
+              GBytes                 *tile_data,
+              ShumateGridPosition    *source_position)
+{
+  g_autoptr(GError) error = NULL;
+  RenderJob *job = NULL;
+  TaskData *data = (TaskData *)g_task_get_task_data (task);
+
+  if (data->current_job != NULL)
+    {
+      g_cancellable_cancel (data->current_job->cancellable);
+      data->current_job = NULL;
+    }
+
+  job = g_new0 (RenderJob, 1);
+  job->cancellable = g_cancellable_new ();
+  job->task = g_object_ref (task);
+  job->data = g_bytes_ref (tile_data);
+  job->source_position = *source_position;
+  data->current_job = job;
+
+  /* If the input cancellable is cancelled, stop the render job. */
+  if (g_task_get_cancellable (task) != NULL)
+    {
+      job->cancellable_handle = g_cancellable_connect (
+        g_task_get_cancellable (task),
+        G_CALLBACK (chain_cancel),
+        job, NULL
+      );
+    }
+
+  if (self->thread_pool == NULL)
+    {
+      self->thread_pool = g_thread_pool_new_full (
+        (GFunc)thread_func,
+        NULL,
+        (GDestroyNotify)render_job_unref,
+        g_get_num_processors () - 1,
+        FALSE,
+        &error
+      );
+      if (self->thread_pool == NULL)
+        {
+          g_critical ("Failed to create thread pool: %s", error->message);
+          return FALSE;
+        }
+    }
+
+  if (!g_thread_pool_push (self->thread_pool, job, &error))
+    {
+      g_critical ("Failed to push job to thread pool: %s", error->message);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+
 /**
  * shumate_style_error_quark:
  *
